"""
    In this module you will find multiple configurations for the app. You can override them by specifying a different
    .env file, exporting environment variables properly named, or providing a settings.toml file.

    Remember that .env files should be in the project's root.
"""

from dynaconf import Dynaconf, Validator
import os
import enum

_HOME = os.environ.get("HOME")

class PersistencyManagers(enum.Enum):
    """
        Possible arguments for "PERSISTENCY_MANAGER" setting,
        telling the possible valid variations of persistency managers.
        # Variations:
            - `SQLITE` = the default SQLite based persistency manager
            - `GCLOUD`  = use a google cloud based persistency manager, measurements are stored in cloud
            - `USER` = Use the user defined persistency manager, specified by the USER_PERSISTENCY_MANAGER_PATH setting.
    """

    SQLITE: str = "SQLITE"
    GCLOUD: str  = "GCLOUD"  
    USER: str = "USER"

settings = Dynaconf(
    envvar_prefix="C4V",
    settings_files=["./settings.toml", "config/.secrets.toml"],
    load_dotenv=True,
    validators=[
        Validator("DATE_FORMAT", default="%Y-%m-%d %H:%M:%S.%f%z"),
        Validator("C4V_FOLDER", default=os.path.join(_HOME, ".c4v")),
        Validator(
            "LOCAL_SQLITE_DB_NAME", default="c4v_db.sqlite"
        ),  # Path to the local sqlite db file
        Validator("LOCAL_SQLITE_DB", default=os.path.join(_HOME, ".c4v/c4v_db.sqlite")),
        Validator(
            "DEFAULT_BASE_LANGUAGE_MODEL", default="PlanTL-GOB-ES/roberta-base-bne"
        ),  # Base language model for the classifier
        Validator(
            "DEFAULT_LANG_MODEL_MIN_LOSS", default=0.15
        ),  # Minimum acceptable loss for base language model, if the loss is greater, a new training is required
        Validator(
            "EXPERIMENTS_DIR", default=os.path.join(_HOME, ".c4v/experiments/")
        ),  # Default name for experiments folder
        Validator(
            "PERSISTENCY_MANAGER",
            default=PersistencyManagers.SQLITE.value,
            is_in=[x.value for x in PersistencyManagers],
        ),  # Specify which persistency manager to use, possible options defined by the PersistencyManagers enum class.
        Validator(
            "USER_PERSISTENCY_MANAGER_PATH", default=None
        ),  # Path for the cusmtom user persistency manager python file. Should export a function: `get_persistency_manager : () -> BasePersistencyManager`
        Validator(
            "USER_PERSISTENCY_MANAGER_MODULE", default=None
        ),  # Module from the imported file where to find the the function
        Validator(
            "CLI_LOGGING_LEVEL", default=10
            # Logging level:
            #   - 10: error, warn, success, info
            #   - 9:  error, warn, success
            #   - 8:  error, warn
            #   - 7:  error
            #   - [1-6]:  reserved for future use
            #   - 0: No logging 
        ),  
        Validator(
            "STORAGE_BUCKET", default=None,
            # Used to choose the name of the bucket where data like the classifier models is stored
        ),
        Validator(
            "SCRAPED_DATA_TABLE", default=None,
            # Used to select the table where to retrieve data from in google cloud
        ),
        Validator(
            "SCRAPING_CLOUD_URL_TRIGGER", default=None
            # Url to use to trigger a scaping process in a google cloud function
        ),
        Validator(
            "CRAWLING_CLOUD_URL_TRIGGER", default=None
            # Url to use to trigger a crawling process in a google cloud function
        ),
        Validator(
            "CLASSIFY_CLOUD_URL_TRIGGER", default=None
            # Url to use to trigger a classification process in a google cloud function
        ),
        Validator(
            "GCLOUD_PROJECT_ID", default=None
            # project id for the gcloud project, where all the cloud operations will be performed
        ),
        Validator(
            "GCLOUD_MAX_CONTENT_LEN", default=6000
        )
    ]
)




"""
    State for the application and functions used to retrieve data needed to feed the 
    page. This is basically a wrapper over c4v-py.
    Also, you have multiple backends to perform operations both locally and in the cloud
"""
# Local imports
from typing import Callable, List
import c4v.microscope as ms
from c4v.microscope.metadata import Metadata
from c4v.config import PersistencyManagers, settings
from c4v.scraper.scraped_data_classes.scraped_data import ServiceClassificationLabels, RelevanceClassificationLabels

# Python imports
from pathlib import Path
from typing import Dict, Any
from urllib import request, parse
import os

# Third party imports
import pandas as pd


class App:
    """
    Main application object, mainly used to get the data to show in the front end
    """

    # Valid options for labels
    relevance_options: List[str] = [
        "ANY",
        *RelevanceClassificationLabels.labels(),
        "NO LABEL",
    ]

    # Service options
    service_options : List[str] = [
        "ANY",
        *ServiceClassificationLabels.labels(),
        "NO LABEL"
    ]

    # Valid options for scraped
    scraped_options: List[str] = ["Any", "Yes", "No"]

    def __init__(self, manager : ms.Manager = None) -> None:

        self._manager = manager or ms.Manager.from_default()

    @property
    def manager(self) -> ms.Manager:
        """
        Manager currently in use
        """
        return self._manager

    def get_dashboard_data(
        self,
        max_rows: int = 100,
        max_content_len: int = 200,
        label_relevance: str = "ANY",
        label_service: str = "ANY",
        scraped: str = "Any",
    ) -> pd.DataFrame:
        """
        Get data to show in the dashboard
        # Parameters:
            - max_rows : `int` = (optional) maximum amount of rows to display
            - max_content_len : `int` = (optional) maximum amount of chars to show in the content field which might be long
            - label : `str` = (optional) description of the label that every instance should have
            - scraped : `str` = (optional) if the instances should be scraped.
        """
        # Some sanity check
        assert label_relevance in App.relevance_options, f"invalid label: {label_relevance}"
        assert label_service in App.service_options, f"invalid label: {label_service}"
        assert scraped in App.scraped_options, f"invalid scraped option: {scraped}"

        # get value depending on if the instances should be scraped
        opt_2_val = {"Any": None, "Yes": True, "No": False}

        query = self._manager.get_all(scraped=opt_2_val[scraped])

        # Add filtering
        if label_relevance == "NO LABEL":
            query = (x for x in query if not x.label_relevance)
        elif label_relevance != "ANY":
            query = (x for x in query if x.label_relevance and x.label_relevance.value == label_relevance)
        
        # Add filtering
        if label_service == "NO LABEL":
            query = (x for x in query if not x.label_service)
        elif label_service != "ANY":
            query = (x for x in query if x.label_service and x.label_service.value == label_service)

        elems = []
        for d in query:

            # In case of null data:
            # Reformat enum fields
            if d.source:
                d.source = d.source.value
                
            if d.label_relevance:
                d.label_relevance = d.label_relevance.value
            
            if d.label_service:
                d.label_service = d.label_service.value

            # Default to empty string
            d.content = d.content or ""

            # truncate body if needed
            content_len = len(d.content)
            d.content = (
                d.content
                if content_len < max_content_len
                else d.content[:max_content_len] + "..."
            )
            elems.append(d)
            print(d.last_scraped)
            # d.last_scraped = None
            # break if gathered enough rows
            if len(elems) == max_rows:
                break
        
        return pd.DataFrame(elems)

    @property
    def available_branchs(self) -> List[str]:
        """
        List of available branches
        """
        # TODO add a function to classifier object to get available experiments and branches for an experiment
        experiments_path = Path(self._manager.local_files_path, "experiments")
        return [str(x.name) for x in experiments_path.glob("*")]

    def available_experiments_for_branch(self, branch_name: str) -> List[str]:
        """
        List of available experiments for a given branch. Raise an error if invalid branch name is provided.

        # Parameters
            - branch_name : `str` = branch whose experiments are to be retrieved
        # Return
            List of experiments corresponding to the given branch
        """
        assert branch_name in self.available_branchs

        # TODO add function to the classifier object to get experiments from branch
        experiments_path = Path(
            self._manager.local_files_path, "experiments", branch_name
        )

        return [str(x.name) for x in experiments_path.glob("*")]

    def experiment_summary(self, branch_name: str, experiment_name: str) -> str:
        """
        Summary for the given experiment defined by its branch name and experiment name. Might be None if
        no summary is found
        # Parameters
            - branch_name : `str ` = Branch name for experiment
            - experiment_name : `str ` = Experiment name for experiment
        # Return
            Summary for the given experiment, or None if not found
        """
        # Sanity check
        assert branch_name in self.available_branchs
        assert experiment_name in self.available_experiments_for_branch(branch_name)

        # TODO Crear funcion en c4v manager que traiga el summary
        summary_path = Path(
            self._manager.local_files_path,
            "experiments",
            branch_name,
            experiment_name,
            "summary.txt",
        )

        # Return None if not exists
        if not summary_path.exists():
            return None

        return summary_path.read_text()

    def classify(self, branch_name: str, experiment_name: str, limit: int = -1, type : str = "relevance"):
        """
        Run a classification process.
        # Parameters
            - branch_name : `str ` = Branch name for experiment
            - experiment_name : `str ` = Experiment name for experiment
            - limit  : `int` = (optional) Max ammount of rows to classify, provide a negative number for no limit
            - type : `str` = (optional) type of classification to perform. One of the following 
                + relevance (default)
                + service
        """
        self._manager.run_pending_classification_from_experiment(
            branch_name, experiment_name, limit=limit, type=type
        )

    def crawl(
        self,
        crawlers_to_use: List[str],
        limit: int,
        progress_function: Callable[[List[str]], None],
    ):
        """
        Run a crawling process
        # Parameters
            - crawlers_to_use : `[str]` = names of the crawlers to use during this crawling process
            - limit : `int` = max amount of rows to store. If limit < 0, no limit is assumed
            - progress_function : `[str] -> None` = Function to call whenever a new bulk of urls is received, intended to be used to add a progress bar
        """
        self._manager.crawl_and_process_new_urls_for(
            crawlers_to_use,
            post_process=progress_function,
            limit=limit,
            save_to_db=True,
        )

    def scrape(self, limit: int) -> int:
        """
        Run a scraping process, scraping only pending rows up to the given limit
        # Parameters
            - limit : `int` = max amount of rows to scrape. If `limit < 0`, 'no limit' is assumed
        # Return
            Exit code for the scraping sub process, 0 if success, anything else for error
        """
        # No pude programar esto de forma que llamara a la función scrape del manejador principal, porque la arquitectura de scrapy
        # demanda estar en el el thread principal, cosa que no pasa con streamlit, así que esto fue lo mejor que pude hacer
        assert isinstance(limit, int)
        return os.system(f"c4v scrape --limit {limit}")

    def upload_model_of_type(self, experiment: str, branch: str, type: str):
        """
        Upload a classifier model
        """
        self.manager.upload_model(branch,experiment, type)

    def download_model_of_type(self, path: str, type: str):
        """
        Download the model of type 'type' to path 'path'
        """
        self.manager.download_model_to_directory(path, type)

    @classmethod
    def cloud_backend():
        """
            Create a cloud backed version of this object
        """
        return CloudApp()

class CloudApp(App):
    """
        App implementation based on cloud actions. Will override 
        operations like scraping, crawling and classifying to perform a cloud request 
    """
    def __init__(self) -> None:

        metadata = Metadata(persistency_manager=PersistencyManagers.GCLOUD.value)
        manager = ms.Manager.from_default(metadata=metadata)
        super().__init__(manager=manager)

    def classify(self, branch_name: str, experiment_name: str, type : str = "relevance", limit: int = -1):
        result = self._make_request(settings.classify_cloud_url_trigger, {"type" : type, "limit" : limit})
    
    def scrape(self, limit: int) -> int:
        result = self._make_request(settings.scraping_cloud_url_trigger, {"limit" : limit})
        # Return response code
        return 0 if result.get("status") == "success" else 1
    
    def crawl(self, crawlers_to_use: List[str], limit: int, progress_function: Callable[[List[str]], None]):
        
        # Perform request
        result = self._make_request(settings.crawling_cloud_url_trigger, {"crawler_names" : crawlers_to_use, "limit" : limit})
        progress_function(["" for _ in range(result.get("crawled", 0))])
    
    def _make_request(self, url : str, data : Dict[str, Any]) -> Dict[str, Any]:
        """
            Perform a post request to the url "url" providing the data in "data"
            and return retrieved answer as dict
        """
        try:
            import google.auth.transport.requests as google_requests
            import google.oauth2.id_token as google_token
        except ImportError as e:
            raise ImportError(f"Could not import google cloud related dependencies. Maybe you're missing the 'gcloud' installation profile?. Error: {e}")

        import json

        # Create post request and encode data
        req = request.Request(url, method= "POST", 
            data = 
                str(
                    json.dumps(data)
                ).encode("utf-8")
            )
        
        # Get auth token 
        auth_req = google_requests.Request()
        id_token = google_token.fetch_id_token(auth_req, url)

        # set proper headers
        req.add_header("Authorization", f"Bearer {id_token}")
        req.add_header('content-type', 'application/json')
        
        # Get response 
        response = request.urlopen(req)
        result = response.read()

        # Parse result
        return json.loads(result)


"""
    The ScrapedData class is a target scheme for scraped data output, it describes how should stored in the db
"""

# Python imports
from dataclasses import dataclass, asdict, field, fields

from typing import List, Dict, Any, Type
from datetime import datetime
import json
import enum

# Local imports
from c4v.config import settings

class LabelSet(enum.Enum):
    """
        Interface for sets of labels that can be attached to to a model for classification
    """

    @classmethod
    def num_labels(cls) -> int:
        """
            Ammount of labels to use during classification. 
            This may be needed in case that the measurement set is using some 
            special labels that are not used for classification
        """
        return len(list(cls.get_id2label_dict().items()))

    @classmethod
    def labels(cls) -> List[str]:
        """
            A list with every possible label string for this labelset
        """
        return [x.name for x in cls]

    @classmethod
    def get_id2label_dict(cls) -> Dict[int, str]:
        """
            Get a dict mapping from ids to a str, representing the labels for this label set
        """
        raise NotImplementedError("Should implement abstract method get_id2label_dict")

    @classmethod
    def get_label2id_dict(cls) -> Dict[str, int]:
        """
            Get a dict mapping labels to int ids, representing the the ids of each label in this labelset
        """
        id2label = cls.get_id2label_dict()
        return { v:k for (k,v) in id2label.items() }

class RelevanceClassificationLabels(LabelSet):
    """
        Labels for Binary classification, telling if a data instance is relevant or not
    """
    IRRELEVANTE: str = "IRRELEVANTE"
    DENUNCIA_FALTA_DEL_SERVICIO: str = "PROBLEMA DEL SERVICIO"
    UNKNOWN: str = "UNKNOWN"

    @classmethod
    def get_id2label_dict(cls) -> Dict[int, str]:
        return {
            0: cls.IRRELEVANTE.value,
            1: cls.DENUNCIA_FALTA_DEL_SERVICIO.value,
        }

class ServiceClassificationLabels(LabelSet):
    """
        Labels for service classification, "electricity", "water", "internet"... 
    """
    AGUA : str = "AGUA"
    ASEO_URBANO : str = "ASEO URBANO"
    COMBINACION : str = "COMBINACIÓN"
    ELECTRICIDAD : str = "ELECTRICIDAD"
    GAS_DOMESTICO : str = "GAS DOMÉSTICO"
    TELECOMUNICACIONES : str = "TELECOMUNICACIONES"
    NO_SERVICIO : str = "NO ES SERVICIO"
    UNKNOWN : str = "UNKNOWN"

    @classmethod
    def get_id2label_dict(cls) -> Dict[int, str]:
        return {
            0 : cls.AGUA.value,
            1 : cls.ASEO_URBANO.value,
            2 : cls.COMBINACION.value,
            3 : cls.ELECTRICIDAD.value,
            4 : cls.GAS_DOMESTICO.value,
            5 : cls.TELECOMUNICACIONES.value,
            6 : cls.NO_SERVICIO.value,
            7 : cls.UNKNOWN.value,
        }
    

class Sources(enum.Enum):
    """
        Every possible source
    """

    UNKOWN: str = "UNKNOWN"
    SCRAPING: str = "SCRAPING"
    CLIENT: str = "CLIENT"

@dataclass()
class ScrapedData:
    """
        This is a general data format class, 
        every data format for other scrapers could have 
        additional fields according to its needs and
        scrapable data, but then they should be able to 
        convert themselves into this format, possibly leaving a 
        few fields as None. Thus, we can be able to 
        easily map from a scrapers's output to a database 
        scheme
    """

    url: str
    last_scraped: datetime = None
    title: str = None
    content: str = None
    author: str = None
    categories: List[str] = field(default_factory=list)
    date: str = None
    label_relevance: RelevanceClassificationLabels = None
    label_service: ServiceClassificationLabels = None
    source: Sources = None

    def pretty_print(self, max_content_len: int = -1) -> str:
        """
            Return a human-readable representation of this data.
            Truncate content if requested
        """

        # create categories string
        categories = "".join(map(lambda s: f"\t+ {s}\n", self.categories))

        max_content_len = max_content_len if max_content_len > 0 else len(self.content)

        # create body content:
        content = self.content
        if max_content_len < len(self.content):
            content = content[:max_content_len] + "..."

        return f"title: {self.title}\nauthor: {self.author}\ndate: {self.date}\ncategories:\n{categories}\nlabel: {self.label_relevance}\nsource: {self.source}\ncontent:\n\t{content}"

    def __hash__(self) -> int:
        return (
            self.url,
            self.last_scraped,
            self.title,
            self.content,
            self.author,
            self.date,
            self.label_relevance,
            self.label_service,
            self.source,
        ).__hash__()

    @property
    def is_scraped(self) -> bool:
        """
            Tells if this instance was already scraped at some point
        """
        return self.last_scraped != None

    def to_dict(self) -> Dict[str, Any]:
        """
            Transform this instance to a dict
        """
        d = asdict(self)

        # Convert label relevance to str
        label_relevance : RelevanceClassificationLabels = d.get("label_relevance")
        if label_relevance:
            d['label_relevance'] = label_relevance.value

        # Convert label service to str
        label_service : ServiceClassificationLabels = d.get("label_service")
        if label_service:
            d['label_service'] = label_service.value
        
        # Convert source to str
        source : Sources = d.get("source")
        if source:
           d['source'] = source.value
        
        # Convert date to str
        last_scraped : datetime = d.get('last_scraped')
        if last_scraped:
            d['last_scraped'] = datetime.strftime(last_scraped, settings.date_format)

        return d

    @classmethod
    def from_dict(cls, scraped_data : Dict[str, Any] ):
        """
            Create scraped data instance from a dict
        """
        # Sanity check
        valid_fields = [x.name for x in fields(cls)]
        if any(k for k in scraped_data.keys() if k not in valid_fields):
            raise ValueError(f"Invalid scraped data dict representation. Invalid fields: {[k for k in scraped_data.keys() if k not in valid_fields]}")

        # Parse relevance label
        label_relevance : str = scraped_data.get("label_relevance")
        if label_relevance:
            scraped_data['label_relevance'] = RelevanceClassificationLabels(label_relevance)
        
        # Parse service label
        label_service : str = scraped_data.get("label_service")
        if label_service:
            scraped_data['label_service'] = ServiceClassificationLabels(label_service)

        # Parse source
        source : str  = scraped_data.get("source")
        if source:
           scraped_data['source'] = Sources(source)
        
        # Parse date 
        last_scraped : str = scraped_data.get('last_scraped')
        if last_scraped:
            scraped_data['last_scraped'] = datetime.strptime(last_scraped, settings.date_format)
        
        return cls(**scraped_data)

class ScrapedDataEncoder(json.JSONEncoder):
    """
        Encoder to turn this file into json format
    """

    def default(self, obj: ScrapedData) -> Dict[str, Any]:
        if isinstance(obj, ScrapedData):
            return asdict(obj)
        elif isinstance(obj, datetime):
            return datetime.strftime(obj, settings.date_format)

        return json.JSONEncoder.default(self, obj)



"""
    Use this script to create a dataset for binary classification for a set of valid datasets 
    (One with data of type A and another with type B, for example)
    Remember that binary classification could mean assign one of two 
    labels, or one of a set of labels.
    # Parameters
        column : `str` = column to use as label, it will be checkef to be present in every dataset
        datasets : `[str]` = list of datasets to merge into a single one, they should be already well formatted
    # Return 
        A new dataset will be created with the concatenation of the provided datasets
"""
# Python imports
import sys
from datetime import datetime
from pytz import tzinfo
from pathlib import Path

# Third party imports
import pandas as pd

# Not enough arguments checking
if len(sys.argv) < 3:
    print("Not enough arguments: expected column and at the least one dataset", file=sys.stderr)
    exit(1)


# get value for column
label_column_name = sys.argv[1]

# get value for datasets 
datasets_files = sys.argv[2:]

# Validate datasets
for file in datasets_files:
    p = Path(file)
    if not p.exists():
        print(f"Invalid file: {file}", file=sys.stderr)

# read csvs
dfs = []
for file in datasets_files:
    print(f"Openning file: {file}...")
    with open(file) as f:
        new_df: pd.DataFrame = pd.read_csv(f)
    print(f"Rows in file: {len(new_df)}")

    # Check that it provides the label column
    if label_column_name not in new_df.columns:
        print(f"Dataset '{file}' does not provides the requested column: {label_column_name}", file=sys.stderr)
        exit(1)

    dfs.append(new_df)

# concat csvs
df = pd.concat(dfs)
del dfs

# Shuffle dataset
df = df.sample(frac=1).reset_index(drop=True)

# time string
date_suffix = datetime.strftime(datetime.now(), "%Y%m%d%H%M%S")
filename = f"train_dataset_{date_suffix}.csv"

print("Saving file to ", filename)
df.to_csv(filename, index=False)



from c4v.classifier.classifier import Classifier
from c4v.classifier.classifier_experiment import (
    ClassifierExperiment,
    ClassifierArgs,
    ClassifierSummary,
)
from c4v.classifier.experiment import (
    BaseExperiment,
    BaseExperimentArguments,
    BaseExperimentSummary,
)



"""
    Format columns of the client dataset in such a way that they match the ScrapedData format, plus a label field:

    url
    last_scraped
    title
    content
    author
    categories
    date
    source
    label_service
    label_relevance

    # Parameters 
    - client_dataset_file : `str` = filename of the client dataset

    --------------------------------------------------------
    Note that this script won't check for duplicate labels:
    
    key  | a | b | c | label          Key | a | d | label 
    ------------------------          --------------------
    1    | _ | _ | _ | l1             1   | _ | _ | l1
    1    | _ | _ | _ | l2       --->  1   | _ | _ | l2
    2    | _ | _ | _ | l3             2   | _ | _ | l3
    2    | _ | _ | _ | l3
"""
# Local imports 
from c4v.scraper.scraped_data_classes.scraped_data import Sources, RelevanceClassificationLabels

# Python imports
import sys
import os
import datetime
import pytz

# Third party
import pandas as pd

# Get command line argument
if len(sys.argv) < 2:
    print("Missing csv filename argument", file=sys.stderr)
    exit(1)

csv_file_name = sys.argv[1]  # name of csv file
with open(csv_file_name) as file:
    df: pd.DataFrame = pd.read_csv(file)

# Rename columns
print("Columns are: ", df.columns)

df.rename(
    {
        "text": "content",
        "tipo_de_evento": "label",
        "tags": "categories",
        "Link de la noticia ": "url",
        "tipo de evento": "label_relevance",
        "servicio_resumido" : "label_service"
    },
    axis="columns",
    inplace=True,
)

# Valid columns
columns = ["url", "title", "content", "author", "categories", "date", "label_relevance", "label_service"]

# Remove irrelevant columns
columns_to_remove = [col for col in df.columns if col not in columns]
print(f"Removing columns: {columns_to_remove}")

df.drop(columns_to_remove, inplace=True, axis=1)
print(f"Columns: {df.columns}")

# Adding source column
source_col_val = Sources.CLIENT.value
print(f"Adding 'source' column with value: {source_col_val}")
df['source'] = source_col_val

# Update label relevance to be relevant for every instance
df["label_relevance"] = RelevanceClassificationLabels.DENUNCIA_FALTA_DEL_SERVICIO.value

# Complete missing columns:
df["last_scraped"] = datetime.datetime.now(tz=pytz.utc)

filename = f"cleaned_{os.path.basename(csv_file_name)}"
print(f"Saving file to: {filename}")
df.to_csv(filename, index=False)



"""
    Module with some helper functions for the manager class
"""
# Local imports
import pathlib
from typing import Type
from c4v.scraper.persistency_manager.base_persistency_manager import (
    BasePersistencyManager,
)

# Python imports
from pathlib import Path
import importlib.util


def _load_user_manager(module_name: str, path: str) -> BasePersistencyManager:
    """
        Loads the user defined manager given the filepath of its python code.
        # Parameters
            - module_name : str = name of the module beeing imported
            - path : `str` = path to python file containing module to import
        # Raises
            - `FileNotFoundError` on file not existing
            - `AttributeError` on `get_persistency_manager` function not found
            - `TypeError` when the returned persistency manager is not a valid implementation of the `BasePersistencyManager` class

    """

    # Error checking
    if not Path(path).exists():
        raise FileNotFoundError(f"{path} is not a valid path, it does not exists")

    spec = importlib.util.spec_from_file_location(module_name, path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    # Try to retrieve persistency manager
    try:
        persistency_manager = module.get_persistency_manager()
    except AttributeError as e:
        raise AttributeError(
            f"module {module} in path {path} does not provides a 'get_persistency_manager : () -> BasePersistencyManager' function. "
            + f"You should do so in order to set up your custom persistency manager object. Note that such object should implement the BasePersistencyManager base class. "
            + f"\n\tError: {e}"
        )

    # Some static checking for the new persistency manager object
    if not isinstance(persistency_manager, BasePersistencyManager):
        raise TypeError(
            f"Object returned by the `get_persistency_manager` doesn't seems to be a valid persistency manager implementation"
        )

    # Return obtained manager
    return persistency_manager



#%%
"""
    Cloud storage manager object, to easily retrieve and upload files from and to google cloud storage
"""
# Internal imports
from c4v.cloud.utils import make_tarfile

# External imports
from google.cloud import storage
import enum

# Python imports
from io import BytesIO
from typing import List
import tempfile
import tarfile
import pathlib
import logging

class ClassifierType(enum.Enum):
    """
        Possible types of classifiers
    """
    RELEVANCE : str = "relevance"
    SERVICE   : str = "service"

    @classmethod
    def choices(cls) -> List[str]:
        return [x.value for x in cls]

class GCSStorageManager:
    """
    Manage data stored un Google Cloud Storage. It gives you access
    to the classifier to use during classifications
    
    # Parameters:
        - bucket : `str` = Name of the bucket to interact with.
        - bucket_prefix_path : `str` = Path inside the provided bucket where files will be looked for
    """

    def __init__(self, bucket_name: str, bucket_prefix_path: str):
        self._client = storage.Client()
        self._bucket = self._client.bucket(bucket_name)
        self._bucket_prefix_path = bucket_prefix_path

    @property
    def bucket(self) -> storage.Bucket:
        """
            Name of the bucket to interact with
        """
        return self._bucket

    @property
    def bucket_prefix_path(self) -> str:
        """
            Path inside the provided bucket where files will be looked for
        """
        return self._bucket_prefix_path

    def download_classifier_model_to(self, type : ClassifierType, filepath : str):
        """
            Try to download the classifier model of type 'type' to the provided filepath in local storage

            # Parameters:
                - type : `ClassifierType` = type of classifier to get
                - filepath : `str` = where to store it in local storage
        """

        # Create name of file, use the prefix path (the path where to look for files) 
        # and find the file as a tar file
        blob_name = f"{self.bucket_prefix_path}/{type.value}.tar"

        # Get blob and download it to local storage
        blob = self.bucket.get_blob(blob_name)

        # Create a templfile where to download the tar file
        with tempfile.NamedTemporaryFile() as file:
            logging.info(f"Retrieving file: {blob_name} from bucket {self.bucket.name}")
            blob.download_to_file(file)

            # untar that file to the desired path
            logging.info(f"Extracting tar file content to '{filepath}'...")
            tar = tarfile.TarFile(file.name)
            tar.extractall(filepath)

    def upload_classifier_model_from(self, type : ClassifierType, filepath : str):
        """
            Try to upload the classifier model of type 'type' stored in the provided filepath in local storage

            # Parameters:
                - type : `ClassifierType` = type of classifier to get
                - filepath : `str` = where to find it in local storage
        """
        # Check if the file exists beforehand 
        if not pathlib.Path(filepath).exists():
            raise ValueError(f"Path in filepath '{filepath}' does not exists")

        # Create name of file, use the prefix path (the path where to look for files) 
        # and find the file as a tar file
        blob_name = f"{self.bucket_prefix_path}/{type.value}.tar"
        logging.info(f"Creating tar file for file: {filepath}...")

        # Create a temp file where to tar the folder
        with tempfile.NamedTemporaryFile() as file:
            make_tarfile(file.name, filepath)
            logging.info(f"Uploading to blob: '{blob_name}' from local storage '{filepath}' to bucket: '{self.bucket.name}'")
            # Get blob and upload to it from local storage
            blob = self.bucket.get_blob(blob_name)
            blob.upload_from_filename(file.name)

    def get_byte_stream(self, filepath: str) -> BytesIO:
        """
            Download a file to bytestream object
            # Parameters:
                - filepath : `str` = path to file inside the bucket to download
            # Return:
                A BytesIO object holding the entirely downloaded object 
        """
        blob_name = filepath if filepath.startswith(self.bucket_prefix_path) \
            else f"{self.bucket_prefix_path}/{filepath}"
        logging.info(f"Bucket {self.bucket.name}: Retrieving file: {blob_name}")
        blob = self.bucket.get_blob(blob_name)
        byte_stream = BytesIO()
        blob.download_to_file(byte_stream)
        byte_stream.seek(0)
        return byte_stream

    def save_file(self, destination_file : str, source_file : str):
        """
            Save a file into the given path 
            # Parameters:
                - destination_file : `str` = path inside the bucket where to store the uploaded file
                - source_file : `str` = path in local storage to the file that will be uploaded
        """
        blob_name = f"{self.bucket_prefix_path}/{destination_file}"
        logging.info(f"Bucket {self.bucket.name}: Saving file with name: {blob_name}")
        blob = self.bucket.blob(blob_name)
        blob.upload_from_filename(source_file)

    def list_files(self, prefix_path: str = None, delimiter: str = None) -> List[str]:
        """
            List files inside the bucket under the given path
            # Parameters:
                - prefix_path : `str` = (optional) path where the files will be looked for; relative to this object's prefix path
            # Return:
                List of filenames of stored objects
        """
        full_prefix = self.bucket_prefix_path if prefix_path is None else f"{self.bucket_prefix_path}/{prefix_path}"
        logging.info(f"Bucket {self.bucket.name}: Listing files with prefix: {full_prefix}")
        blobs = self._client.list_blobs(self.bucket, prefix=full_prefix, delimiter=delimiter)
        return [blob.name for blob in blobs]



"""
    Utility functions for the cloud module
"""
import tarfile
import os.path

def make_tarfile(output_filename : str, source_dir : str):
    """
        Create a tar file from the given directory to the specified output file
        # Parameters
            - output_filename : `str` = Name of the file to be written
            - source_dir : `str` = path to the directory in local storage to be tared
    """
    with tarfile.open(output_filename, "w") as tar:
        # this arcname arguments prevents from this tar to include the local filesystem layout
        tar.add(source_dir, arcname=os.path.basename(source_dir))




"""
    The manager may need to store some persistent configuration data
    between runs, here we find the scheme requested for the system
"""
# Local imports
from typing import Type
from c4v.config import settings

# Python imports
import dataclasses
import json


@dataclasses.dataclass
class Metadata:
    """
        Data required and stored by the manager to perform common operations.    
        # Parameters:
            - classifier_model : `str` = Default classifier model. Could be a huggingface model or a locally stored one
            - base_language_model : `str` = Default base language model used as a base for training the new classifier model
            - persistency_manager : `str` = Persistency manager to use. One of the possible variants of the c4v.config.PersistencyManagers enum class
            - user_persistency_manager_path : `str` = Path to a file returning a custom persistency manager to use for the CLI
                                                    tool. Such file should provide a function `get_persistency_manager : () -> BasePersistencyManager` 
                                                    that will be used to retrieve the Persistency Manager object to use.    
                                                    If not provided, the default sqlite based manager is used. 
    """

    classifier_model: str = settings.default_base_language_model  # Absolute path for the model to load
    base_language_model: str = settings.default_base_language_model  # Absolute path for the base model to load
    persistency_manager: str = settings.persistency_manager  # Type of manager to use
    user_persistency_manager_path: str = settings.user_persistency_manager_path  # Path to a persistency manager to use in the CLI tool. If not provided, defaults to a SQLite one
    user_persistency_manager_module: str = settings.user_persistency_manager_module  # module where to get the persistency manager itself

    def to_json_str(self, pretty: bool = False) -> str:
        """
            Return a json-formated string representation for this object
        """
        indent = 1 if pretty else None
        return json.dumps(self, cls=_MetadataJSONEncoder, indent=indent)

    @classmethod
    def from_json_str(cls, json_str: str):
        """
            Create instance from json-formated string
        """
        return cls(**json.loads(json_str))

    def to_json(self, filename: str, pretty: bool = False) -> str:
        """
            Return a json-formated string representation for this object
        """
        indent = 1 if pretty else None
        with open(filename, "w") as f:
            return json.dump(self, f, cls=_MetadataJSONEncoder, indent=indent)

    @classmethod
    def from_json(cls, json_path: str):
        """
            Create an instance from a json file
            Parameters:
                json_path : str = path to file to load, may raise FileNotFound error if file does not exists
        """
        with open(json_path) as f:
            try:
                return cls(**json.load(f))
            except TypeError as e:  # in case of an unexpected argument
                raise TypeError(f"Couldn't parse metadata object from json. Error: {e}")


class _MetadataJSONEncoder(json.JSONEncoder):
    """
        Class to serialize metadata into a json file
    """

    def default(self, o):
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)
        return super().default(o)



"""
    This script will trim out all but the specified columns of a dataset specified by its filename, 
    and store the result in another file. 
    If you want to create a regular dataset with a specified label field
    # Arguments
        dataset : `str` = filename of the dataset whose rows will be removed 
        -l label_name : `str` = You can optionally provide a field name after the flag -l to mark it as a label field, so that
                                it can be used as label in a classification pipeline. If present, it should be right before
                                the column list
        columns : `[str]` = list of names of columns that will be kept, they should be 
                            valid columns and an error will be raised if not
    # Return
        A new file will be written in the current directory with the new dataset.
"""

# Python imports
import sys
from pathlib import Path

# Third party imports
import pandas as pd

# -- < Input & sanity check > ---------------------------

# Utility function that will log an error and end the program  execution with a non-zero exit code
def raise_error(msg : str):
    print(f"[ERROR] {msg}", file=sys.stderr)
    exit(1)


# should provide at the least name of this file, path to dataset, and at the least one column to remove
if len(sys.argv) < 3: 
    raise_error("Expected at the least 2 arguments, path to dataset as csv and a non-empty list of columns to remove")

# Try to parse label flag
if sys.argv[2] == '-l' and len(sys.argv) < 5:
    raise_error("Expected at the least 3 arguments, path to dataset as csv, label flag value and a non-empty list of columns to remove")
elif sys.argv[2] == "-l":
    label_column = sys.argv[3]
    columns_start_index = 4
else:
    label_column = None
    columns_start_index = 2

print("label column is:", label_column)

# Parse file and columns
file = sys.argv[1]
columns = sys.argv[columns_start_index:]

# Check that the given file is a valid one
path_to_file = Path(file)

if not path_to_file.exists():
    raise_error("The given file does not exists")
elif not file.endswith(".csv"):
    raise_error("This doesn't seems to be a valid csv file")

# Load dataset
dataframe = pd.read_csv(file)

# Get dataset columns
actual_columns = list(dataframe.columns)

# Check consistent columns 
if any(x not in actual_columns for x in columns) or ( False if not label_column else (label_column not in actual_columns)):
    raise_error("Inconsistent columns and csv. Provided columns not in csv format")

# -- < filter out undesired columns > ---------------------
print("Deleting columns...")
columns_to_drop = [c for c in actual_columns if c not in columns]
dataframe.drop(columns_to_drop, axis='columns', inplace=True)

# Rename column to label if requested so
if label_column:
    print(f"Renaming column {label_column} to 'label'...")
    dataframe.rename({label_column : 'label'}, axis='columns', inplace=True)

# -- < Save to file > -------------------------------------
filename = path_to_file.stem + "_trimmed_cols.csv"
print(f"Saving new csv to: {filename}...")

dataframe.to_csv(filename, index=False)



"""
    This file implements a class that manages local storage 
    using regular files.
"""
# Python
from typing import Any, Iterator, Dict, List, Tuple, Set
import dataclasses
import datetime

# Local imports
from c4v.scraper.persistency_manager.base_persistency_manager import (
    BasePersistencyManager,
)
from c4v.scraper.scraped_data_classes.scraped_data import ServiceClassificationLabels, RelevanceClassificationLabels, ScrapedData, Sources
from c4v.config import settings
import uuid

# Third party imports
from google.cloud import bigquery
import firebase_admin
from firebase_admin import credentials
from firebase_admin import firestore
from c4v.scraper.scraped_data_classes.scraped_data import ScrapedData
import datetime
import pytz



DATE_FORMAT = settings.date_format


class BigQueryManager(BasePersistencyManager):
    """
        Use Big Query based storage 
        to manage persistency.

        Recently added and fill-pending measurements are stored in firestore, 
        while complete data is stored in big query, as it is more reliable 
        for massive ammounts of data.

        Big query is read and insert only for this manager object.

        # Parameters 
        - table_name : `str` = name of the big query table (prefixed with dataset name) where to perform operations 
                                (you can fill it with the C4V_SCRAPED_DATA_TABLE env variable or setting variable)
        - bq_client : `bigquery.Client` = client object to use 
        - project_id : `str` = name of the project in gcloud to connect to in order to access firestore
                                (you can fill it by using the C4V_GCLOUD_PROJECT_ID variable)
    """                         

    _firestore_db = None

    def __init__(self, table_name : str, bq_client : bigquery.Client, project_num : str, gcloud_max_content_len : int = settings.gcloud_max_content_len):
        self._table_name = table_name
        self._client = bq_client
        self._project_num = project_num
        self._gcloud_max_content_len = gcloud_max_content_len
        self._init_firestore()

    @property
    def table_name(self) -> str:
        return self._table_name

    @property
    def client(self) -> bigquery.Client:
        return self._client

    @property
    def project_num(self) -> str:
        return self._project_num

    @property
    def _firestore_scraped_data_table_name(self) -> str:
        return u"scraped_data"

    @property 
    def _firestore_db_ref(self) -> Any:
        return self._firestore_db.collection(self._firestore_scraped_data_table_name)

    @property
    def bq_date_format(self) -> str:
        # Required to send data to big query on list of json 
        return  "%Y-%m-%d %H:%M:%S.%f"
    
    @property
    def gcloud_max_content_len(self) -> int:
        """
            Max size for body len for an article
        """
        return self._gcloud_max_content_len

    def _init_firestore(self):
        """
            Initialization code for firestore
        """
        if (BigQueryManager._firestore_db != None):
            return # nothing to init

        self._credentials = credentials.ApplicationDefault()
        firebase_admin.initialize_app(self._credentials, {
        # 'projectId': self._project_num,
        })
        BigQueryManager._firestore_db = firestore.client()


    def _get_all_from_bq(
        self, limit: int = -1, scraped: bool = None, order_by: List[str] = None
    ) -> Iterator[ScrapedData]:
        """
            Get all data from Big Query according to the get all arguments
        """
        # Retrieve all data stored
        
        # Create the query string
        # Order
        if order_by:
            parsed_orders = " ORDER BY " + ", ".join(
                [
                    f"{name} {order}"
                    for (order, name) in (
                        self._parse_order_and_field_from_order_by_str(order_by=ord)
                        for ord in order_by
                    )
                ]
            )
        else:
            parsed_orders = ""

        # if scraped = true, take only scraped. If false, take non-scraped. If none, take all
        if scraped:
            query = f"SELECT * FROM `{self.table_name}` WHERE last_scraped IS NOT NULL{parsed_orders}"
        elif scraped == False:
            query = f"SELECT * FROM `{self.table_name}` WHERE last_scraped IS NULL{parsed_orders}"
        else:
            query = f"SELECT * FROM `{self.table_name}` {parsed_orders}"

        # If limit less than 0, then take as much as you can
        assert isinstance(limit, int), "limit argument should be a valid integer"
        if limit < 0:
            query += ";"
        else:
            query += f" LIMIT {limit};"

        res = self.client.query(query)
        for row in res:
            # Decompose row
            (url, last_scraped, title, content, author, date, label_relevance, label_service, source, categories) = row

            # Add timezone to date
            if last_scraped:
                last_scraped = pytz.utc.localize(last_scraped)
                
            if label_relevance:
                try:
                    label_relevance = RelevanceClassificationLabels(label_relevance)
                except:  # not a known label
                    label_relevance = RelevanceClassificationLabels.UNKNOWN
            else:
                label_relevance = None

            if label_service:
                try:
                    label_service = ServiceClassificationLabels(label_service)
                except:  # not a known label
                    label_service = ServiceClassificationLabels.UNKNOWN
            else:
                label_service = None

            if source:
                try:
                    source = Sources(source)
                except:  # unknown source
                    source = Sources.UNKOWN

            yield ScrapedData(
                url=url,
                last_scraped=last_scraped,
                title=title,
                content=content,
                author=author,
                date=date,
                categories=categories,
                label_relevance=label_relevance,
                label_service=label_service,
                source=source,
            )
        
    def _get_all_from_firestore(
        self, limit: int = -1, scraped: bool = None, order_by: List[str] = None, ready : bool = None
        ) -> Iterator[ScrapedData]:
        """
            Get all data from Firestore according to the get all arguments.
            Firestore data will have the following format:
                <scraped data fields>
                - was_scraped : bool = if this row is already scraped
                - is_ready_for_bq : bool = if this row is ready to be upgraded to bq
            
            Use the "ready" argument to tell if you want the retrieved data to be ready or if you just want the 
            rows not yet ready or already ready
        """
        # Sanity check
        assert isinstance(limit, int), "limit should be integer"

        # If nothing to retrieve, just finish
        if limit == 0:
            return

        scraped_data_ref = self._firestore_db_ref

        # If scraped specified
        if scraped == True:
            scraped_data_ref = scraped_data_ref.where(u'was_scraped', u'==', True)
        elif scraped == False:
            scraped_data_ref = scraped_data_ref.where(u'was_scraped', u'==', False)
        
        # If ready specified
        if ready != None:
            assert isinstance(ready, bool), "ready argument should be bool"
            scraped_data_ref = scraped_data_ref.where(u'is_ready_for_bq', u'==', ready)

        # if order by is provided:
        if order_by:
            parsed_orders = (self._parse_order_and_field_from_order_by_str(order, True) for order in order_by)
            for (order, field) in parsed_orders:
                scraped_data_ref = scraped_data_ref.order_by(field, direction=order)

        # if limit is provided
        if limit > 0:
            scraped_data_ref = scraped_data_ref.limit(limit)

        return ( self._from_firestore_to_scraped_data(x) for x in scraped_data_ref.stream())
    
    def get_all(self, limit: int = -1, scraped: bool = None, order_by: List[str] = None) -> Iterator[ScrapedData]:
        # Merge data from firestore to bq
        # TODO Create a merge function to merge results in both queries

        returned = 0
        for x in self._get_all_from_firestore(limit, scraped , order_by):
            yield x
            returned += 1

        # Sanity check
        assert returned <= limit or limit < 0, "There should not be more returned urls than the provided limit"

        # Check if already at the max 
        if returned == limit and limit > 0:
            return

        # This is an optimization, there's no non-scraped data in BQ
        if scraped == False: 
            return 

        for x in self._get_all_from_bq(limit, scraped , order_by):
            yield x


    def _from_firestore_to_scraped_data(self, data : Any) -> ScrapedData:
        """
            Used to convert from firestore stored object to actual scraped data object.
            It will remove the firestore-specific fields.
            'data' is an instance object as it comes from firestore
            
        """
        data = data.to_dict()
        del data['was_scraped']
        del data['is_ready_for_bq']

        return ScrapedData.from_dict(data)

    def _from_scraped_data_to_firestore(self, scraped_data : ScrapedData) -> ScrapedData:
        """
            Used to convert from actual scraped data to an actual firestore storable object.
            It will add firestore-specific fields:
                - was_scraped : bool = tells if this instance is already scraped
                - is_ready_for_bq : bool = tells if this instance can be saved into big query
        """
        d = scraped_data.to_dict()
        d['was_scraped'] = scraped_data.is_scraped
        d["is_ready_for_bq"] = scraped_data.label_relevance != None and scraped_data.label_relevance != None and scraped_data.is_scraped

        return d
    
    def _filter_scraped_urls_bq(self, urls: List[str]) -> List[str]:
        return self._filter_known_urls_bq(urls)

    def _filter_scraped_urls_firestore(self, urls: List[str]) -> List[str]:
        # Get table reference
        scraped_data_ref = self._firestore_db.collection(self._firestore_scraped_data_table_name)

        # Max ammount of elements in "in" query of firestore
        chunk_size = 10

        # get urls such that there's not 
        lo = 0
        result = []
        while (lo < len(urls)):
            # Current chunk to check
            chunk = set(urls[lo:lo + chunk_size])

            # Urls that are both in the list and scraped
            scraped_urls = {x.get("url") for x in scraped_data_ref.where("url", 'in', chunk).where("was_scraped", "==", True).stream()}
            
            for url in chunk:
                if url not in scraped_urls:
                    result.append(url)

            # firestore has a max "in" query size of 10
            lo += chunk_size
        
        return result

    def filter_scraped_urls(self, urls: List[str]) -> List[str]:
        # Just filter from firestore what was filtered from bq. 
        # Note that bq might be a more precise tool as it holds only complete data
        return self._filter_scraped_urls_firestore(self._filter_scraped_urls_bq(urls))

    def _filter_known_urls_bq(self, urls: List[str]) -> List[str]:
        # connect to db and check for each url if such url exists, if so, then it's known and scraped.
        # Otherwise it is unkown for the DB, and thus it's also not scraped
        query_str = f"SELECT url FROM `{self.table_name}` WHERE url in UNNEST(@url);"
        job_config = bigquery.QueryJobConfig(
            query_parameters=[
                bigquery.ArrayQueryParameter("url", "STRING", urls)
            ]
        )

        # Create a set to discard results in such set 
        res : Set(str)= set(urls)

        # if last_scraped is null, then it wasn't scraped
        query = self.client.query(query_str, job_config=job_config)
        for row in query:
            (url,) = row
            # Remove urls in DB
            if url in res:
                res.remove(url)
        
        return list(res)

    def _filter_known_urls_firestore(self, urls: List[str]) -> List[str]:
        # Get table reference
        scraped_data_ref = self._firestore_db.collection(self._firestore_scraped_data_table_name)

        # Max ammount of elements in "in" query of firestore
        chunk_size = 10

        # get urls such that there's not 
        lo = 0
        result = []
        while (lo < len(urls)):
            # Current chunk to check
            chunk = urls[lo:lo + chunk_size]

            # Urls that are both in the list and scraped
            known_urls = {x.get("url") for x in scraped_data_ref.where("url", 'in', chunk).stream()}
            
            for url in chunk:
                if url not in known_urls:
                    result.append(url)

            # firestore has a max "in" query size of 10
            lo += chunk_size
        
        return result

    def filter_known_urls(self, urls: List[str]) -> List[str]:
        # return urls that are neither in bq nor in firestore
        return self._filter_known_urls_firestore(self._filter_known_urls_bq(urls))

    def _was_scraped_bq(self, url: str) -> bool:
        # Connect to db and check if object was scraped
        query_str = f"SELECT 1 FROM `{self.table_name}` WHERE url=@url;"
        job_config = bigquery.QueryJobConfig(query_parameters=[
            bigquery.ScalarQueryParameter("url", "STRING", url)
        ])

        for _ in self.client.query(query_str, job_config=job_config):
            return True # consume just one, altought there should be at the most one element in this query
        
        return False

    def _was_scraped_firestore(self, url: str) -> bool:
        # Get reference to table
        scraped_data_ref = self._firestore_db.collection(self._firestore_scraped_data_table_name)

        # check if exists
        for _ in scraped_data_ref.where('url', '==', url).where('was_scraped', '==', True).stream():
            return True
        
        return False

    def was_scraped(self, url: str) -> bool:
        # Just check if it was scraped in some database
        return self._was_scraped_bq(url) or self._was_scraped_firestore(url)

    def _save_bq(self, url_data: List[ScrapedData]):
        # Save data into database
        if not url_data:
            return

        data_to_insert = [dataclasses.asdict(data) for data in url_data]
        for data in data_to_insert:
            label_relevance: RelevanceClassificationLabels = data["label_relevance"]
            data["label_relevance"] = label_relevance.value if label_relevance else label_relevance 
            label_service: ServiceClassificationLabels = data["label_service"]
            data["label_service"] = label_service.value if label_service else label_service 
            source: Sources = data["source"]
            data["source"] = source.value if source else source
            data["last_scraped"] = datetime.datetime.strftime( data["last_scraped"], self.bq_date_format) if data["last_scraped"] else None

            # truncate content if necessary
            content : str = data.get("content")
            if content and len(content) > self.gcloud_max_content_len:
                data['content'] = content[:self.gcloud_max_content_len]
            

        del url_data

        # Delete instances before insert them to give it priority to the ones to be saved right now
        query_str = f"DELETE FROM `{self.table_name}` WHERE url IN @urls;"
        query_config = bigquery.QueryJobConfig(query_parameters=
            [
                bigquery.ArrayQueryParameter("urls", "STRING", [d['url'] for d in data_to_insert])
            ]
        )
        query_job = self.client.query(query_str, job_config=query_config)

        # Now insert updated rows
        errors = self.client.insert_rows_json(self.table_name, data_to_insert)
        assert not errors, f"There was some errors while trying to insert new data to big query: {errors}"

    def _save_firestore(self, url_data: List[ScrapedData]):
        # Perform save operation for firestore

        db_ref = self._firestore_db_ref

        # Convert into firestore-specific format
        fs_data = ((x.url , self._from_scraped_data_to_firestore(x)) for x in url_data)
        for (url, data) in fs_data:
            db_ref.document(self._get_url_uuid(url)).set(data)

    def save(self, url_data: List[ScrapedData]):
        # Note that a regular save operation won't save to BQ, it will go to firestore first
        return self._save_firestore(url_data)

    def _retrieve_and_delete_ready_from_firestore(self) -> List[ScrapedData]:
        """
            This function will try to retrieve all ready rows in DB, and then delete them
            in a transactional manner
        """
        # TODO maybe use transaction?
        data_ref = self._firestore_db_ref
        data_stream = data_ref.where("is_ready_for_bq", '==', True)

        results = []
        for data in data_stream.stream():
            # save each data row and then delete it
            sd = self._from_firestore_to_scraped_data(data)
            results.append(sd)
            data_ref.document(self._get_url_uuid(sd.url)).delete()
        
        return results

    def move(self):
        """
            Move ready data from firestore to big query, and then remove it from 
            firestore
        """
        data_to_move = self._retrieve_and_delete_ready_from_firestore()
        try:
            # try to save data to bq 
            self._save_bq(data_to_move)
        except Exception as e:
            # save it back to firestore if it wasn't possible
            self.save(data_to_move)
            

    def delete(self, urls: List[str]):
        # Bulk delete provided urls
        query_str = f"DELETE FROM `{self.table_name}` WHERE url=@url;"
        for url in urls:
            job_config = bigquery.QueryJobConfig(query_parameters=
                [
                    bigquery.ScalarQueryParameter("url", "STRING", url)
                ]
            )
            self.client.query(query_str, job_config=job_config)

    def _parse_order_and_field_from_order_by_str(
        self, order_by: str, use_firestore_order : bool = False
    ) -> Tuple[str, str]:
        """
            Parse order and name of field from the given str
            # Parameters:
                - order_by : str = A string with the following format:
                    (+|-)<field_name> where field_name is the name of a field in ScrapedData,
                    '+' corresponds to a an ascending order, and '-' corresponds to a descending order
                - use_firestore_order : bool = if should use firestore order literals
            # Return:
                Tuple such that
                ('ASC' | 'DESC', <field_name>) 
        """

        # Check valid format
        if not order_by:
            raise ValueError(
                "Invalid value for order_by string, I'm getting an empty string"
            )
        elif order_by[0] != "-" and order_by[0] != "+":
            raise ValueError(
                f"Invalid value for order_by string, first char should be the order, on of: [-, +]. Provided string: {order_by}"
            )
        elif len(order_by) < 2:
            raise ValueError(
                f"Invalid value for order_by string, name of field not provided. Given string: {order_by}"
            )

        # Parsing order
        if use_firestore_order:
            order = firestore.Query.ASCENDING if order_by[0] == "+" else firestore.Query.DESCENDING
        else:
            order = "ASC" if order_by[0] == "+" else "DESC"

        # Parsing field name
        valid_fields = [d.name for d in dataclasses.fields(ScrapedData)]
        field_name = order_by[1:]

        # Raise error if not a valid string
        if field_name not in valid_fields:
            valid_fields_str = "\n".join([f"\t* {s}\n" for s in valid_fields])
            raise ValueError(
                f"Invalid value for order_by string, provided name doesn't match any field in ScrapedData. Provided field: {field_name}.\nAvailable fields:\n{valid_fields_str}"
            )

        return (order, field_name)

    @staticmethod
    def _get_url_uuid(url : str) -> str:
        """
            Use this function to get an uuid from an url, 
            useful to hash urls into ids in firestore
        """
        return str(uuid.uuid3(uuid.NAMESPACE_URL, url))

def _parse_dict_to_url_data(obj: Dict[str, Any]) -> ScrapedData:
    """
    Parse dict object into a new object instance
    """
    return ScrapedData(**obj)



"""
    Entry point for the streamlit app
"""

# Python imports
import dataclasses
from typing import List
from pathlib import Path
# Third party imports
import streamlit as sl

# Internal imports
import c4v.dashboard.app as app
from c4v.config import settings

sl.write("# 🔬 C4V Microscope")
sl.write(
    "In the following dashboard we will see how the scraping, crawling and classification works."
)

# -- < Back End sub menu > -------------------------------------
backend = sl.sidebar.write("## Select Backend")

sl.sidebar.write(
    """
        Select a backend. If using a local backend, operations like scraping and crawling
        will be performed in you're local computer, and displayed data will come from
        your local sqlite database. 

        If you select a cloud backend, you'll need to set up the appropiate permissions
        and configuration environment variables, and operations will be performed
        on cloud. Displayed data will come from the cloud as well.
    """
    )

backend = sl.sidebar.radio("Backend", ["local", "cloud"])
if backend == "local":
    app = app.App()
elif backend == "cloud":
    app = app.CloudApp()
else:
    raise ValueError(f"Invalid backend option: '{backend}'")

# -- < Filtering > ---------------------------------------------
# Add a title to the sidebar
sl.sidebar.write("## Filters")
sl.sidebar.write("Specify which data you want to see with the following filters")

# Add a selector to filter by label
relevance = sl.sidebar.selectbox(
    "Relevance: ", options=app.relevance_options, help="relevance label assigned by the classifier"
)

service = sl.sidebar.selectbox(
    "Service: ", options=app.service_options, help="service label assigned by the classifier"
)

scraped = sl.sidebar.selectbox(
    "Scraped: ",
    options=app.scraped_options,
    help="Whether the instance data is complete or not",
)

use_max_rows = sl.sidebar.checkbox(
    "Limit Rows: ", value=True, help="Max ammount of rows to show"
)

# -- < Row Limits > -------------------------------------------
# If limiting the max amount of rows, then ask for a max value
if use_max_rows:
    max_rows = sl.sidebar.number_input(
        "Max: ", 0, value=100, help="Maximum amount of rows to retrieve and show"
    )
else:
    max_rows = -1

# -- < Classification controls > ------------------------------
# Help Message
sl.sidebar.write("-----")
sl.sidebar.write("## Classification")
sl.sidebar.write(
    "Run a classification process, select parameters with the following controls."
)

# Classification type
classifier_type = sl.sidebar.selectbox(
    "Type: ",
    ["relevance", "service"], 
    help="The classifier type specifies which information is predicted by the classifier"
)

# Branch names
NO_BRANCH_NAME = "No Branch"
if backend == "local":
    branch_name = sl.sidebar.selectbox(
        "Branch: ",
        [NO_BRANCH_NAME] + app.available_branchs,
        help="Experiment branch to select a classifier model from",
    )
    experiment_options = (
        []
        if branch_name == NO_BRANCH_NAME
        else app.available_experiments_for_branch(branch_name)
    )
    experiment_name = sl.sidebar.selectbox(
        "Experiment: ",
        experiment_options,
        help="Experiment name corresponding to provided branch",
    )

    # Description for experiment if provided
    if experiment_name:
        sl.sidebar.write("### Summary for this experiment: ")
        sl.sidebar.text_area(
            "Summary", app.experiment_summary(branch_name, experiment_name), height=100
        )

# Set limits
use_classiffication_limit = sl.sidebar.checkbox(
    "Limit",
    value=True,
    help="Max ammount of rows to use during classification. Might be useful if you're running out of memory",
)
if use_classiffication_limit:
    max_rows_to_classify = sl.sidebar.number_input(
        "Max rows",
        0,
        value=100,
        help="Max amount of rows to classify. If you want no limit, uncheck the checkbox above",
    )
else:
    max_rows_to_classify = -1

# Run classification button
def run_local_classification_callback():
    if not experiment_name:
        sl.warning(
            "No experiment set up. Select one to choose a classifier model to use during classification"
        )
    else:
        sl.info("Running classification process, this might take a while...")
        try:
            app.classify(branch_name, experiment_name, max_rows_to_classify, type=classifier_type)
            sl.success("Classification finished")
        except MemoryError as e:
            sl.error(
                f"Unable to classify using model {branch_name}/{experiment_name} {('and up to ' + str(max_rows_to_classify)) if max_rows_to_classify >= 0 else ''}.    "
                + f"Error: {e}"
            )

def run_cloud_classification_callback():
    sl.info("Running classification process, this might take a while...")
    try:
        app.classify(classifier_type, max_rows_to_classify)
        sl.success("Classification finished")
    except Exception as e:
        sl.error(
            f"Unable to classify using model of type {classifier_type} {('and up to ' + str(max_rows_to_classify)) if max_rows_to_classify >= 0 else ''}.    "
            + f"Error: {e}"
        )

sl.sidebar.button(
    "Classify",
    help="Perform the classification process",
    on_click=run_local_classification_callback if backend == "local" else run_cloud_classification_callback,
)

# Downloading and uploading
def upload_and_download_model():
    sl.sidebar.write("### Uploading and downloading")
    sl.sidebar.write("""
                Upload a model to be used online during classification. 
                Specify the type of model to tell the purpose of such model
            """)
    # Check that a valid branch is provided
    if not branch_name and not experiment_name:
        sl.sidebar.write("⚠️ Provide branch name and experiment name to upload a model")
        return
    elif not branch_name:
        sl.sidebar.write("⚠️ Provide branch name to upload a model")
        return
    elif not experiment_name:
        sl.sidebar.write("⚠️ Provide experiment name to upload a model")
        return
    
    # Check that there's a bucket where to upload the model
    if not settings.storage_bucket:
        sl.sidebar.write("⚠️ there's no bucket properly configured where to upload the classifier")
        return

    # Switch to tell if it should download or upload
    upload_or_download = sl.sidebar.radio("Upload or Download", ["upload", "download"])

    def upload():
        sl.info(f"Uploading model in {branch_name}/{experiment_name} of type '{classifier_type}' to bucket...")
        try:
            app.upload_model_of_type(experiment_name, branch_name, classifier_type)
            sl.success("Upload successful")
        except Exception as e:
            sl.error(f"Could not complete upload. Error: '{e}'")

    if upload_or_download == "upload":
        sl.sidebar.button(
            "Upload",
            help=f"Upload model in '{experiment_name}/{branch_name}' of type '{classifier_type}' to bucket {settings.storage_bucket or '<no bucket specified in environment>'}",
            on_click=upload
            )
        return
    
    # Ask for a path where to download 
    path = sl.sidebar.text_input("Download Path")
    def download():
        path_obj = Path(path)
        if not path or not path_obj.exists() or not path_obj.is_dir():
            sl.error(f"The path '{path}' is not a valid download path for a model")
            return

        sl.info(f"Downloading model of type' {classifier_type}' to '{path}'...")

        # try to perform download
        try: 
            app.download_model_of_type(path, classifier_type)
            sl.success("Download successful")
        except Exception as e:
            sl.error(f"Could not complete download. Error: '{e}'")


    sl.sidebar.button(
        "Download", 
        help=f"Download model of type '{classifier_type}' from bucket '{settings.storage_bucket}'",
        on_click=download
    )

if backend == "local":
    upload_and_download_model()

# -- < Crawling Controls >  -----------------------------------
sl.sidebar.write("------")
sl.sidebar.write("## Crawling")
sl.sidebar.write("Crawl for new urls in the available sites")

# Limit
use_crawl_limit = sl.sidebar.checkbox(
    "Limit", help="Use a max amount of urls to be crawled at the same time", value=True
)
if use_crawl_limit:
    crawl_limit = sl.sidebar.number_input("Max instances to add", 0, value=100)
else:
    crawl_limit = -1
# Crawlers to use
crawlers = sl.sidebar.multiselect(
    "Crawl for: ",
    options=app.manager.available_crawlers(),
    help="Sites that can be scraped for new data",
)

# utility class to keep count of how much instance were crawled so far
@dataclasses.dataclass
class Count:
    count: int = 0

    def add(self, additional: int):
        self.count += additional


crawled_so_far = Count()

# Function to print a message with the current amount of crawled urls
def generate_progress_infinite():

    place_holder_msg = sl.empty()

    def progress_infinite(urls: List[str]):
        # Display some text telling how much urls have been crawled so far
        crawled_so_far.add(len(urls))
        place_holder_msg.info(f"Crawled {crawled_so_far.count} urls...")

    return progress_infinite


# Function to generate a finite progress bar function
def generate_progress_finite():
    sl.write("Progress")
    progress_bar = sl.progress(0)
    progress_msg = sl.empty()
    progress_msg.markdown(f"Crawled {crawled_so_far.count}/{crawl_limit}...")

    def progress_finite(urls: List[str]):
        crawled_so_far.add(len(urls))
        progress_msg.markdown(f"Crawled {crawled_so_far.count}/{crawl_limit}...")
        progress_bar.progress(int(100 * (crawled_so_far.count / crawl_limit)))

    return progress_finite


def run_crawl_callback():
    """
    Run a classification and show an info tect
    """
    # If nothing to do, just end the process
    if not crawlers or crawl_limit == 0:
        return

    # Start classification process
    sl.info("Starting crawling process")
    app.crawl(
        crawlers,
        crawl_limit,
        generate_progress_finite() if crawl_limit > 0 else generate_progress_infinite(),
    )

    # Tell the user how much urls were crawled
    if crawled_so_far.count == 0:
        sl.warning("No urls were crawled. Maybe there were no urls to crawl?")
        return

    sl.success(f"Succesfully crawled {crawled_so_far.count} elements!")


# Button to run
sl.sidebar.button(
    "Crawl",
    help="Run a crawling limit with the provided limits",
    on_click=run_crawl_callback,
)

# -- < Scraping controllers > ---------------------------------
sl.sidebar.write("------")
sl.sidebar.write("## Scraping")
sl.sidebar.write("Scrape instances pending for scraping")

# Limit
use_scrape_limit = sl.sidebar.checkbox(
    "Limit", help="Use a max amount of urls to be scraped at the same time", value=True
)
if use_scrape_limit:
    scrape_limit = sl.sidebar.number_input("Max instances to scrape", 0, value=100)
else:
    scrape_limit = -1

# Called when crawl button is pressed
def scrape_callback():
    sl.info("Starting scraping process...")
    if app.scrape(scrape_limit) == 0:
        sl.success("Successfull scraping! 🤩")
    else:
        sl.error("Scraping process crashed 😵")


# Scrape button
sl.sidebar.button(
    "Scrape",
    help="run a classification process, trying to scrape data for instances with incomplete data",
    on_click=scrape_callback,
)

# -- < Show Dashboard > ---------------------------------------
sl.dataframe(app.get_dashboard_data(label_relevance=relevance, label_service=service, max_rows=max_rows, scraped=scraped))
